import glob
from os.path import dirname, join
from loguru import logger
import json
import re


def file_paths():
    base_path = dirname(__file__)

    search_pattern = join(base_path, 'raw', '*.txt')

    files = sorted(glob.glob(search_pattern))

    return files


def safe_read(file: str):
    try:
        with open(file) as fp:
            return fp.read()
    except:
        logger.exception(f"Error in reading {file}")
        raise


def safe_json_load(file: str):
    try:
        return json.loads(safe_read(file))
    except:
        logger.exception(f"JSON Decoding error in {file}")
        raise


def cleaner(doc: str):
    return re.sub('\d+\s+[a-zA-ZŞşğĞüÜıİ]+\s+\d{4}\s+PAYLAŞ\s+yorum\s+yaz(\s+a)?', '', doc, flags=re.I)


def load_raw_corpus(return_iter: bool = True, base_path=None, clean=True):
    """Load corpus of sample news tokenized into sentences.

    Examples
    --------
    >>> from sadedegel.dataset import load_raw_corpus
    >>> sents = load_raw_corpus(return_iter=False)
    >>> type(sents[0])
    <class 'str'>

    """
    if base_path is None:
        base_path = dirname(__file__)

    search_pattern = join(base_path, 'raw', '*.txt')

    logger.debug("Search path {}".format(search_pattern))

    files = sorted(glob.glob(search_pattern))

    if return_iter:
        if clean:
            return (cleaner(safe_read(file)) for file in files)
        else:
            return (safe_read(file) for file in files)
    else:
        if clean:
            return [cleaner(safe_read(file)) for file in files]
        else:
            return [safe_read(file) for file in files]


def load_sentence_corpus(return_iter: bool = True, base_path=None):
    """Load corpus of sample news tokenized into sentences.

    Examples
    --------
    >>> from sadedegel.dataset import load_sentence_corpus
    >>> sents = load_sentence_corpus(return_iter=False)
    >>> type(sents[0])
    <class 'dict'>

    >>> len(sents[0]['sentences'])
    62
    """
    if base_path is None:
        base_path = dirname(__file__)

    search_pattern = join(base_path, 'sents', '*.json')

    logger.debug("Search path {}".format(search_pattern))

    files = sorted(glob.glob(search_pattern))

    if return_iter:
        return map(safe_json_load, files)
    else:
        return [safe_json_load(file) for file in files]



import click
from .core import load


@click.command
def cli():
    pass


@cli.command()
@click.argument('doc')
def summarize(doc):
    summarizer = load()

    summary = summarizer(doc)

    click.echo(summary)


if __name__ == '__main__':
    summarize()  # pylint: disable=E1120



from enum import Enum
from .tokenize import RegexpSentenceTokenizer
from .summarize import FirstK


class TokenizerEnum(Enum):
    AUTO_TOKENIZER = 0


def tr_tokenizer():
    from transformers import AutoTokenizer  # pylint: disable=import-outside-toplevel

    return AutoTokenizer.from_pretrained("dbmdz/bert-base-turkish-cased")


class NLPPipeline:
    def __init__(self, sent_tokenizer=RegexpSentenceTokenizer(),
                 tokenizer=TokenizerEnum.AUTO_TOKENIZER,
                 summarizer=FirstK(0.5)):
        self.sent_tok = sent_tokenizer
        self.summarizer = summarizer

        if isinstance(summarizer, FirstK):
            self.tokenizer = None
        else:
            if tokenizer == TokenizerEnum.AUTO_TOKENIZER:
                self.tokenizer = tr_tokenizer()

    def __call__(self, doc):
        sents = self.sent_tok(doc)

        if self.tokenizer is None:
            summary = self.summarizer(sents)
        else:
            raise NotImplementedError("Summarization with a word tokenizer is not there yet.")

        return summary


class Summarizer:
    def __init__(self):
        self.pipeline = NLPPipeline()

    def __call__(self, doc):
        return self.pipeline(doc)


def load():
    """Extraction-based document summarizer pipeline

     Examples
     --------

     >>> from sadedegel import load
     >>> from sadedegel.dataset import load_raw_corpus
     >>> summarizer = load()
     >>> raw = load_raw_corpus(False)

     >>> news_summary = summarizer(raw[2])

     >>> len(raw[2])
     3179

     Default text summarizer only takes first half of the all sentences in document.
     >>> sum((len(sent) for sent in news_summary))
     1699

     """
    return Summarizer()



import re
from joblib import load
from pathlib import Path
from os.path import dirname
from loguru import logger
import re
from typing import List

from .ml.sbd import load_model

__tr_upper__ = "ABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZ"
__tr_lower__ = "abcçdefgğhıijklmnoöprsştuüvyz"

__tr_lower_abbrv__ = ['hz.', 'dr.', 'prof.', 'doç.', 'org.', 'sn.', 'st.', 'mah.', 'mh.', 'sok.', 'sk.', 'alb.', 'gen.',
                      'av.', 'ist.', 'ank.', 'izm.', 'm.ö.', 'k.k.t.c.']


def tr_lower(s: str) -> str:
    return s.replace("I", "ı").replace("İ", "i").lower()


def tr_upper(s: str) -> str:
    return s.replace("i", "İ").upper()


def flatten(l2: List[List]):
    flat = []
    for l in l2:
        for e in l:
            flat.append(e)

    return flat


class Span:
    def __init__(self, i: int, span, doc):
        self.doc = doc
        self.i = i
        self.value = span

    def __str__(self):
        return self.value

    @property
    def text(self):
        return self.doc.raw[slice(*self.value)]

    def span_features(self):
        is_first_span = self.i == 0
        is_last_span = self.i == len(self.doc.spans) - 1
        word = self.doc.raw[slice(*self.value)]

        if not is_first_span:
            word_m1 = self.doc.raw[slice(*self.doc.spans[self.i - 1].value)]
        else:
            word_m1 = '<D>'

        if not is_last_span:
            word_p1 = self.doc.raw[slice(*self.doc.spans[self.i + 1].value)]
        else:
            word_p1 = '</D>'

        features = {}

        # All upper features
        if word_m1.isupper() and not is_first_span:
            features["PREV_ALL_CAPS"] = True

        if word.isupper():
            features["ALL_CAPS"] = True

        if word_p1.isupper() and not is_last_span:
            features["NEXT_ALL_CAPS"] = True

        # All lower features
        if word_m1.islower() and not is_first_span:
            features["PREV_ALL_LOWER"] = True

        if word.islower():
            features["ALL_LOWER"] = True

        if word_p1.islower() and not is_last_span:
            features["NEXT_ALL_LOWER"] = True

        # Century
        if tr_lower(word_p1).startswith("yüzyıl") and not is_last_span:
            features["NEXT_WORD_CENTURY"] = True

        # Number
        if (tr_lower(word_p1).startswith("milyar") or tr_lower(word_p1).startswith("milyon") or tr_lower(
                word_p1).startswith("bin")) and not is_last_span:
            features["NEXT_WORD_NUMBER"] = True

        # Percentage
        if tr_lower(word_m1).startswith("yüzde") and not is_first_span:
            features["PREV_WORD_PERCENTAGE"] = True

        # In parenthesis feature
        if word[0] == '(' and word[-1] == ')':
            features["IN_PARENTHESIS"] = True

        # Suffix features
        m = re.search(r'\W+$', word)

        if m:
            subspan = m.span()
            suff2 = word[slice(*subspan)][:2]

            if all((c in '!\').?’”":;…') for c in suff2):
                features['NON-ALNUM SUFFIX2'] = suff2

        if word_p1[0] in '-*◊' and not is_last_span:
            features["NEXT_BULLETIN"] = True

        # title features
        # if word_m1.istitle() and not is_first_span:
        #     features["PREV_TITLE"] = 1

        """NOTE: 'Beşiktaş’ın' is not title by python :/
        """
        m = re.search(r'\w+', word)

        if m:
            subspan = m.span()
            if word[slice(*subspan)].istitle():
                features["TITLE"] = True

        m = re.search(r'\w+', word_p1)

        if m:
            subspan = m.span()

            if word_p1[slice(*subspan)].istitle() and not is_last_span:
                features["NEXT_TITLE"] = True

        # suffix features
        for name, symbol in zip(['DOT', 'ELLIPSES', 'ELLIPSES', 'EXCLAMATION', 'QUESTION'],
                                ['.', '...', '…', '?', '!']):
            if word.endswith(symbol):
                features[f"ENDS_WITH_{name}"] = True

        # prefix abbreviation features
        if '.' in word:

            if any(tr_lower(word).startswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features[f"PREFIX_IS_ABBRV"] = True
            if any(tr_lower(word).endswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features[f"SUFFIX_IS_ABBRV"] = True
            else:
                prefix = word.split('.', maxsplit=1)[0]
                features["PREFIX_IS_DIGIT"] = prefix.isdigit()

        # if '.' in word_m1:
        #     prefix_m1 = word_m1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_m1) in __tr_lower_abbrv__ and not is_first_span:
        #         features[f"PREV_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["PREV_PREFIX_IS_DIGIT"] = prefix_m1.isdigit()
        #
        # if '.' in word_p1:
        #     prefix_p1 = word_p1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_p1) in __tr_lower_abbrv__ and not is_last_span:
        #         features[f"NEXT_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["NEXT_PREFIX_IS_DIGIT"] = prefix_p1.isdigit()

        return features


def is_eos(span: Span, sentences: List[str]) -> int:
    start = 0
    eos = []
    for s in sentences:
        idx = span.doc.raw.find(s, start) + len(s) - 1
        eos.append(idx)

        start = idx

    b, e = span.value

    for idx in eos:
        if b <= idx <= e:
            return 1

    return 0


class Doc:
    sbd = None

    def __init__(self, raw: str):
        if Doc.sbd is None:
            Doc.sbd = load_model()

        self.raw = raw

        _spans = [match.span() for match in re.finditer(r"\S+", self.raw)]

        self.spans = [Span(i, span, self) for i, span in enumerate(_spans)]

        y_pred = Doc.sbd.predict((span.span_features() for span in self.spans))

        # logger.info(y_pred)

        eos_list = [end for (start, end), y in zip(_spans, y_pred) if y == 1]

        self.sents = []

        for i, eos in enumerate(eos_list):
            if i == 0:
                self.sents.append(self.raw[:eos].strip())
            else:
                self.sents.append(self.raw[eos_list[i - 1] + 1:eos].strip())



from math import ceil
from random import sample


class BasicSummarizer:
    def __call__(self, sents, return_iter=False):
        if return_iter:
            return self._select(sents)
        else:
            return list(self._select(sents))


class FirstK(BasicSummarizer):
    def __init__(self, k=3):
        self.k = k
        super().__init__()

    def _select(self, sents):

        if type(self.k) == int:
            limit = self.k
        else:
            limit = min(ceil(self.k * len(sents)), len(sents))

        for i, sent in enumerate(sents):
            if i < limit:
                yield sent


class RandomK(BasicSummarizer):
    def __init__(self, k=3):
        self.k = k
        super().__init__()

    def _select(self, sents):

        if type(self.k) == int:
            limit = self.k
        else:
            limit = min(ceil(self.k * len(sents)), len(sents))

        yield from sample(sents, limit)



from fastapi.openapi.utils import get_openapi

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from pydantic.typing import List
from sadedegel.tokenize import Doc
from math import ceil
from sadedegel.summarize import RandomK, FirstK
import click
from sadedegel.about import __version__


class Document(BaseModel):
    text: str


class Summary(BaseModel):
    sentences: List[str]
    length: int
    ratio: float
    summary_length: int


app = FastAPI()


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="sadedeGel",
        version=__version__,
        description="sadedeGel is an extraction-based summarizer for Turkish news content",
        routes=app.routes,
    )
    openapi_schema["info"]["x-logo"] = {
        "url": "https://avatars0.githubusercontent.com/u/2204565?s=280&v=2"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

app.add_middleware(
    CORSMiddleware,
    allow_origins='*',
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.post("/sadedegel/random")
async def random(doc: Document, ratio: float = 0.2):
    """Baseline Random summarizer.
        Returning random K sentence of the document.
    """
    sentences = Doc(doc.text).sents

    summary_length = max(1, ceil(len(sentences) * ratio))

    return Summary(sentences=RandomK(summary_length)(sentences), length=len(sentences), summary_length=summary_length,
                   ratio=ratio)


@app.post("/sadedegel/firstk")
async def firstk(doc: Document, ratio: float = 0.2):
    """Baseline FirstK summarizer.
    Returning first K sentence of the document
    """
    sentences = Doc(doc.text).sents

    summary_length = max(1, ceil(len(sentences) * ratio))

    return Summary(sentences=FirstK(summary_length)(sentences), length=len(sentences), summary_length=summary_length,
                   ratio=ratio)


@click.command()
@click.option("--host", '-h', help="Hostname", default="0.0.0.0")  # nosec
@click.option("--log-level",
              type=click.Choice(['debug', 'info'], case_sensitive=False), help="Logging Level", default="info")
@click.option("--reload", is_flag=True, default=False, help="enable/disable auto reload for development.")
@click.option("--port", help="Port", default=8000)
def server(host, log_level, reload, port):
    """Span a sadedeGel http server instance."""

    if reload:
        uvicorn.run("sadedegel.server.__main__:app", host=host, log_level=log_level, reload=reload, port=port)
    else:
        uvicorn.run(app, host=host, log_level=log_level, reload=reload, port=port)


if __name__ == "__main__":
    server()
